#! /usr/env/python
# -*- coding: utf-8 -*-

from pyjess import Template
from pyjess import Jess
from pyjess import Molecule
from pathlib import Path
import os
import biotite.structure.io.pdbx as pdbx
import biotite.structure.io.pdb as pdb
import warnings

def run_jess(structure_files: str, distance: float, 
            templates = Path(__file__).parent / "resources" / "data" / "template_structures") -> list:
    """
    
        Runs Jess using stored isopeptide bond templates (default templates). 
        Only intrachain matches are considered. In case of multiple models 
        (delimited by ENDMDL) only the first is considered.
        If .cif files are found they are converted into .pdb models (and stored in "./converted_models/").

        Args:
            structure_files (list): containing paths to .cif and .pdb files
            distance (float): used to set rmsd_threshold, distance_cutoff, max_allowed_distance jess parameters
                               higher values will cause a more permissive search. Note that this does not influence
                               probability prediction
            templates (str): directory containing templates to be used with .pdb extension

        Returns:
            list of pyjess._jess.Hit 

        Rises:
            FileNotFoundError if template files are not found

    """
    template_files = [str(p.resolve()) for p in Path(templates).glob("*.pdb")]
    if not template_files:
        raise FileNotFoundError("Jess templates not found.")

    pdb_files = [p for p in structure_files if p[-4:]==".pdb"]
    cif_files = [p for p in structure_files if p[-4:]==".cif"]

    #if not pdb_files+cif_files:
    #    raise FileNotFoundError("No PDB/CIF files found in the specified directory.")
    
    # Convert
    if cif_files:
        warnings.warn("CIF files detected. Converting them into PDB format.", UserWarning)
        converted_models_dir = os.path.join(os.getcwd(), "converted_models")
        os.makedirs(converted_models_dir, exist_ok=True)
        for cif_file_path in cif_files:
            cif_file = pdbx.CIFFile.read(cif_file_path)
            arr = pdbx.get_structure(cif_file)
            pdb_file = pdb.PDBFile()
            pdb_file.set_structure(arr)
            pdb_file_path = os.path.join(converted_models_dir, os.path.basename(cif_file_path).replace(".cif", ".pdb"))
            pdb_file.write(pdb_file_path)
            pdb_files.append(pdb_file_path)

    rmsd_threshold, distance_cutoff, max_dynamic_distance = [distance]*3
    
    # Load templates
    templates = []
    for path in list(template_files):
        templates.append(Template.load(path, id=os.path.basename(path).split(".pdb")[0]))
    jess = Jess(templates)
    
    # Run on structures
    hits = []
    if pdb_files:
        for path in pdb_files:
            mol = Molecule.load(path, id=path)
            query = jess.query(mol, rmsd_threshold=rmsd_threshold, 
                                distance_cutoff=distance_cutoff, 
                                max_dynamic_distance=max_dynamic_distance)
            hits.extend(query)

    return hits


#! /usr/env/python
# -*- coding: utf-8 -*-

import biotite.structure as struc
from isopeptor.constants import MEAN_BOND_LENGTH
from isopeptor.constants import STD_DEV_BOND_LENGTH
import warnings
import numpy as np

def get_bond_length(structure:struc.AtomArray, chain:str, r1_bond:int, 
                    r2_bond:int, r2_bond_name:str):
    """

        Get bond length

        Args:
            structure (biotite.structure.AtomArray)
            chain (str)
            r1_bond (int)
            r2_bond (int)
            r2_bond_name (str)

    """
    bond_length = None
    n_atom = None
    c_atom = None

    # Get N atom (from lys or asn)
    n_res  = [atom for atom in structure if atom.res_id == r1_bond and atom.chain_id == chain and atom.atom_name == "NZ"]
    if len(n_res) == 0:
        n_res = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "ND2"]
    if len(n_res) == 0:
        warnings.warn("Missing nitrogen from bond. Bond length not calculated", UserWarning)
    else:
        n_atom = n_res[0]

    if r2_bond_name == "ASN" or r2_bond_name == "ASP":
        # CG is the last
        c_res = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CG"]

    if r2_bond_name == "GLU":
        # CD is the last
        c_res = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CD"]
    
    if len(c_res) == 0:
        warnings.warn("Missing carbon from bond. Bond length not calculated", UserWarning)
    else:
        c_atom = c_res[0]
    
    if n_atom != None and c_atom != None:
        bond_length = round(struc.distance(n_atom, c_atom), 3)

    return bond_length

def get_bond_stats(structure:struc.AtomArray, chain:str, r1_bond:int, 
                    r2_bond:int, r2_bond_name:str) -> tuple:
    """

        Get bond zscore

        Args:
            structure (biotite.structure.AtomArray)
            chain (str)
            r1_bond (int)
            r2_bond (int)
            r2_bond_name (str)

        Returns:
            (bond_length:float, bond_length_zscore:float, bond_length_allowed:bool)

    """
    # Get bond length
    bond_length = get_bond_length(structure, chain, r1_bond, r2_bond, r2_bond_name)
    bond_length_zscore = None
    if bond_length != None:
        bond_length_zscore = round((bond_length - MEAN_BOND_LENGTH)/STD_DEV_BOND_LENGTH, 3)
        bond_length_allowed = False
        if np.abs(bond_length_zscore) < 4:
            bond_length_allowed = True

    return (bond_length, bond_length_zscore, bond_length_allowed)




#! /usr/env/python
# -*- coding: utf-8 -*-

class BondElement:
    """

        Stores isopeptide bond predictions

        Attributes:
            struct_file (str): path to structure file
            protein_name (str): protein name from file name
            rmsd (float): rmsd with template (in A)
            template (str): template name
            chain (str): protein chain
            r1_bond (int): residue number of residue 1 involved in bond
            r_cat (int): residue number of catalytic residue
            r2_bond (int): residue number of residue 2 involved in bond

    """
    def __init__(
        self, struct_file: str, protein_name: str, rmsd: float, template: str,
        chain: str, r1_bond: int, r_cat: int, r2_bond: int,
        r1_bond_name: str, r_cat_name: str, r2_bond_name: str
        ):
        self.struct_file = struct_file
        self.protein_name = protein_name
        self.rmsd = rmsd
        self.template = template
        self.chain = chain
        self.r1_bond = r1_bond
        self.r_cat = r_cat
        self.r2_bond = r2_bond
        self.r1_bond_name = r1_bond_name
        self.r_cat_name = r_cat_name
        self.r2_bond_name = r2_bond_name
        self.r_asa: float | None = None
        self.probability: float | None = None
        self.bond_length: float | None = None
        self.bond_length_zscore: float | None = None
        self.bond_length_allowed: bool | None = None
        self.pseudo_phi: float | None = None
        self.pseudo_psi: float | None = None
        self.pseudo_omega: float | None = None
        self.phi_psi_likelihood: float | None = None
        self.phi_psi_allowed: bool | None = None
        self.omega_psi_likelihood: float | None = None
        self.omega_psi_allowed: bool | None = None
        self.omega_phi_likelihood: float | None = None
        self.omega_phi_allowed: bool | None = None

    def __repr__(self):
        s = f"BondElement(struct_file={self.struct_file}, protein_name={self.protein_name}, rmsd={self.rmsd}, template={self.template}, "+\
            f"chain={self.chain}, r1_bond={self.r1_bond}, r_cat={self.r_cat}, r2_bond={self.r2_bond}, "+\
            f"r1_bond_name={self.r1_bond_name}, r_cat_name={self.r_cat_name}, r2_bond_name={self.r2_bond_name}, "+\
            f"bond_type={self.bond_type}, r_asa={self.r_asa}, probability={self.probability})"
        return s



#! /usr/env/python
# -*- coding: utf-8 -*-

import biotite.structure.io.pdb as pdb
import biotite.structure as struc
import os
from isopeptor.structure import get_structure

def get_structure_asa(pdb_file_path: str) -> tuple:
    """

        Return structure and ASA

        Args:
            pdb_file_path (str): path to PDB structure

    """
    # Load structure
    structure = get_structure(pdb_file_path)
    # Calc sasa
    structure_sasa = struc.sasa(structure, point_number=500)
    
    return (structure_sasa, structure)


#! /usr/env/python
# -*- coding: utf-8 -*-

import joblib
import numpy as np
import os
from pathlib import Path
#import warnings
#import warnings
#warnings.filterwarnings("ignore")

CLF = None

def get_model():
    global CLF
    if CLF is None:
        model_path = Path(__file__).parent / "resources" / "models" / "logistic_regression.pkl"
        if not os.path.isfile(model_path):
            raise ValueError("Model file not found.")
        CLF = joblib.load(model_path)
    return CLF

def predict(rmsd:float, r_asa:float)->float:
    """

        Predicts isopeptide bond probability using rmsd and r_asa

        Args:
            rmsd (float): rmsd with template
            r_asa (float): relative solvent accessible surface

    """
    clf = get_model()
    prob_isopep = clf.predict_proba(np.array([[rmsd, r_asa]]))[:,1]
    return round(prob_isopep[0], 3)


#! /usr/env/python
# -*- coding: utf-8 -*-

import biotite.structure as struc
import numpy as np
from isopeptor.constants import DIHEDRAL_ANGLES_TRHESHOLDS
from pathlib import Path
import warnings
import os
import joblib

KDE = None
def get_dihedral_distrib_models():
    global KDE
    bond_types = ["CnaA-like", "CnaB-like"]
    bond_pairs = ["phi_psi", "omega_phi", "omega_psi"]
    if KDE is None:
        KDE = {bond_type:
                {bond_pair: None for bond_pair in bond_pairs
                } for bond_type in bond_types}

        for bond_type in bond_types:
            for bond_pair in bond_pairs:
                model_path = Path(__file__).parent / "resources" / "models" / f"{bond_type}_{bond_pair}.pkl"
                if not os.path.isfile(model_path):
                    raise ValueError(f"Model file not found in {model_path}.")
                KDE[bond_type][bond_pair] = joblib.load(model_path)

    return KDE

def get_dihedral_angles(structure:struc.AtomArray, chain:str, r1_bond:int, 
                    r2_bond:int, r2_bond_name:str) -> tuple:
    """

        Get dihedral torsion angles. Returns pseudo_omega, pseudo_psi and pseudo_phi in a tuple

        Args:
        - structure: biotite.structure.AtomArray
        - chain: str
        - r1_bond: int
        - r2_bond: int
        - r2_bond_name: str

        Returns:
        - (pseudo_omega, pseudo_psi, pseudo_phi)

    """
    
    lys_cd = [atom for atom in structure if atom.res_id == r1_bond and atom.chain_id == chain and atom.atom_name == "CD"]
    lys_ce = [atom for atom in structure if atom.res_id == r1_bond and atom.chain_id == chain and atom.atom_name == "CE"]
    lys_nz = [atom for atom in structure if atom.res_id == r1_bond and atom.chain_id == chain and atom.atom_name == "NZ"]
    asn_nz = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "ND2"]

    if len(lys_cd) > 0:
        lys_cd = lys_cd[0]
    else:
        lys_cd = None

    if len(lys_ce) > 0:
        lys_ce = lys_ce[0]
    else:
        lys_ce = None
    
    if len(lys_nz) > 0:
        n_atom = lys_nz[0]
    elif len(asn_nz) > 0:
        n_atom = asn_nz[0]
    else:
        n_atom = None

    if r2_bond_name == "ASN" or r2_bond_name == "ASP":
        # CG is the last
        c1 = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CG"]
        if len(c1) > 0:
            c1 = c1[0]
        else: 
            c1 = None
        # CB
        c2 = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CB"]
        if len(c2) > 0:
            c2 = c2[0]
        else:
            c2 = None
        # CA
        c3 = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CA"]
        if len(c3) > 0:
            c3 = c3[0]
        else:
            c3 = None

    if r2_bond_name == "GLU":
        # CD is the last
        c1 = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CD"]
        if len(c1) > 0:
            c1 = c1[0]
        else:
            c1 = None
        # CG
        c2 = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CG"]
        if len(c2) > 0:
            c2 = c2[0]
        else:
            c2 = None
        # CB
        c3 = [atom for atom in structure if atom.res_id == r2_bond and atom.chain_id == chain and atom.atom_name == "CB"]
        if len(c3) > 0:
            c3 = c3[0]
        else:
            c3 = None

    if any([True if atom == None else False for atom in [lys_cd, lys_ce, n_atom, c1, c2, c3]]):
        pseudo_omega, pseudo_psi, pseudo_phi = [None]*3
        warnings.warn("Missing atom from bond. Dihedrals not calculated", UserWarning)
        #print([lys_cd, lys_ce, n_atom, c1, c2, c3])
    else:
        # This corresponds to the peptide omega angle (torsion angle on bond CN)
        pseudo_omega = struc.dihedral(c2, c1, n_atom, lys_ce)
        pseudo_omega = round(pseudo_omega*180/np.pi, 3)
        # This corresponds to psi (one carbon is used instead of a second N)
        pseudo_psi = struc.dihedral(c3, c2, c1, n_atom)
        pseudo_psi = round(pseudo_psi*180/np.pi, 3)
        # This corresponds to phi (lys_cd should be bound to oxygen and then to Nitrogrn to form the next peptide bond)
        pseudo_phi = struc.dihedral(c1, n_atom, lys_ce, lys_cd)
        pseudo_phi = round(pseudo_phi*180/np.pi, 3)
    
    return (pseudo_omega, pseudo_psi, pseudo_phi)

def get_dihedral_angles_stats(structure:struc.AtomArray, chain:str,
                                r1_bond:int, r2_bond:int, r2_bond_name:str, bond_type:str) -> tuple:
    """

    Get dihedral torsion angles and stats about their distribution

    Args:
        structure (biotite.structure.AtomArray)
        chain (str)
        r1_bond (int)
        r2_bond (int)
        r2_bond_name (str)
        bond_type (str)

    Returns:
            (
                pseudo_omega:float, pseudo_phi:float, pseudo_psi:float, phi_psi_likelihood:float, omega_psi_likelihood:float, 
                omega_phi_likelihood:float, phi_psi_allowed:bool, omega_psi_allowed:bool, omega_phi_allowed:bool
            )

    """
    pseudo_omega, pseudo_psi, pseudo_phi = get_dihedral_angles(structure, chain, r1_bond, 
                    r2_bond, r2_bond_name)
    if pseudo_omega == None or pseudo_psi == None or pseudo_phi == None:
        phi_psi_likelihood, omega_psi_likelihood, omega_phi_likelihood, phi_psi_allowed, omega_psi_allowed, omega_phi_allowed = [None]*6
    else:
        # Load models
        kde = get_dihedral_distrib_models()

        # Get likelihoods
        phi_psi_likelihood = round(kde[bond_type]["phi_psi"].score([[pseudo_phi, pseudo_psi]]), 3)
        omega_psi_likelihood = round(kde[bond_type]["omega_psi"].score([[pseudo_omega, pseudo_psi]]), 3)
        omega_phi_likelihood = round(kde[bond_type]["omega_phi"].score([[pseudo_omega, pseudo_phi]]), 3)

        phi_psi_allowed, omega_psi_allowed, omega_phi_allowed = [False]*3
        
        if phi_psi_likelihood > DIHEDRAL_ANGLES_TRHESHOLDS[bond_type]["phi_psi"]:
            phi_psi_allowed = True

        if omega_psi_likelihood > DIHEDRAL_ANGLES_TRHESHOLDS[bond_type]["omega_psi"]:
            omega_psi_allowed = True

        if omega_phi_likelihood > DIHEDRAL_ANGLES_TRHESHOLDS[bond_type]["omega_phi"]:
            omega_phi_allowed = True

    return (
                pseudo_omega, pseudo_phi, pseudo_psi, phi_psi_likelihood, omega_psi_likelihood, omega_phi_likelihood, 
                phi_psi_allowed, omega_psi_allowed, omega_phi_allowed
            )




#! /usr/env/python
# -*- coding: utf-8 -*-

"""

    Run isopeptide bond prediction from command line.

"""

from isopeptor.isopeptide import Isopeptide 
import argparse

parser = argparse.ArgumentParser(
    prog = "isopeptor",
    description = "Run isopeptide bond prediction from command line. Usage: isopeptor path/to pdb files/ > isopeptide_bonds.csv"
)

parser.add_argument(
    "path_to_structure_files", 
    help="Path to directory containing .pdb/.cif files.", 
    type=str
)

parser.add_argument(
    "--distance", 
    required=False, 
    help="Specifies permissivity of jess search. The higher, the more permissive.", 
    type=float,
    default=3
)

parser.add_argument(
    "--fixed_r_asa", 
    required=False, 
    help="Fixes the relative solvent accessible area using a value between 0 and 1 to speed up the prediction.", 
    type=float,
    default=None
)

parser.add_argument(
    "--eval_geometry", 
    required=False, 
    help="Run geometric evaluation of isopeptide bonds.", 
    action='store_true'
)

def main():
    args = parser.parse_args()
    i = Isopeptide(args.path_to_structure_files, args.distance, args.fixed_r_asa)
    i.predict()
    if args.eval_geometry:
        i.get_geometry()
    i.print_tabular()

if __name__ == "__main__":
    main()


#! /usr/env/python
# -*- coding: utf-8 -*-

import biotite.structure.io.pdb as pdb
import biotite.structure as struc
import os

def get_structure(pdb_file_path: str) -> tuple:
    """

        Load structure and remove hydrogens and hetero atoms

        Args:
            pdb_file_path (str): path to input pdb file

        Rises:
             FileNotFoundError if pdb_file_path not found

    """
    if not os.path.isfile(pdb_file_path):
        raise FileNotFoundError("PDB file not found.")
    pdb_file = pdb.PDBFile.read(pdb_file_path)
    # Load structure
    structure = struc.array([atom for atom in pdb_file.get_structure()[0]])
    # Exlude hetero atoms and hydrogens
    structure = structure[(structure.hetero==False) & (structure.element != "H")]

    return structure


#! /usr/env/python
# -*- coding: utf-8 -*-

import re
from typing import List
import warnings
import numpy as np
import os
from pathlib import Path
from isopeptor.jess_wrapper import run_jess
from isopeptor.asa import get_structure_asa
from isopeptor.bond import BondElement 
from isopeptor.constants import MAX_ASA
from isopeptor.constants import BOND_TYPE
from isopeptor.logistic_regression import predict
from isopeptor.structure import get_structure
from isopeptor.bond_length import get_bond_stats
from isopeptor.dihedrals import get_dihedral_angles_stats

class Isopeptide:
    """
    
        Handles isopeptide bond prediction running pyjess via the jess_wrapper module
        and solvent accessible area via the asa module. Stores isopeptide bond predictions
        as a list of BondElement. Prediction is run for all structures from struct_dir. 
        Structures in .pdb format are directly analysed. If structures are in .cif format, they are first converted into .pdb.
        If multiple matches with an isopeptide bond signature are detected, only the one
        with the lowest RMSD is retained.

        Attributes:
            struct_dir (str): where pdb/cif files are located
            distance (float): that specifies permissivity of jess search
            jess_output (None | str): that stores jess raw output for debug purposes
            isopeptide_bonds (list): stores isopeptide bonds as BondElement elements
            fixed_r_asa (float | None): which ranges between 0 and 1 that fixes r_asa allowing to skip its calculation

        Example:
            
            >>> # Use a fixed solvent accessible area for a quick prediction:
            >>> from isopeptor.isopeptide import Isopeptide
            >>> i = Isopeptide("tests/data/test_structures", distance=1.5, fixed_r_asa=0.1)
            >>> i.predict()
            >>> i.isopeptide_bonds[0]
            BondElement(struct_file=tests/data/test_structures/8beg.pdb, protein_name=8beg, rmsd=0.0, template=8beg_A_590_636_729, chain=A, r1_bond=590, r_cat=636, r2_bond=729, r1_bond_name=LYS, r_cat_name=ASP, r2_bond_name=ASN, bond_type=CnaA-like, r_asa=0.1, probability=0.991)            
            
            >>> # Calculate solvent accessible area for a more accurate (and slow) prediction:
            >>> i = Isopeptide("tests/data/test_structures", distance=1.5)
            >>> i.predict()
    """
    def __init__(self, struct_dir: str, distance: float = 3, fixed_r_asa: float | None = None):
        """
        
            Raises
               ValueError if fixed_r_asa not between 0 and 1
        
        """
        self.struct_dir: str = struct_dir
        self.distance: float = distance
        self.jess_hits: list | None = None
        self.isopeptide_bonds: List[BondElement] = []
        self.fixed_r_asa: float | None = fixed_r_asa
        if self.fixed_r_asa != None:
            if self.fixed_r_asa < 0 or self.fixed_r_asa > 1:
                raise ValueError(f"fixed_r_asa is not in 0-1 range. Found: {self.fixed_r_asa}")
        pdb_files = [str(p) for p in Path(self.struct_dir).glob("*.pdb")]
        cif_files = [str(p) for p in Path(self.struct_dir).glob("*.cif")]
        self.structure_files = pdb_files + cif_files
        self.geometry_calculated = False

    def predict(self):
        """

            Predict isopeptide bonds by 1. running jess template-based search,
            2. calculating asa, 3. predicting isopeptide bond probability with
            logistic regression.

        """

        self.jess_hits = run_jess(self.structure_files, self.distance)
        if len(self.jess_hits) == 0:
            warnings.warn("No isopeptide bond predictions detected. Try increasing the distance parameter.", UserWarning)
            return
        self._load_hits()
        self._reduce_redundant()
        if self.fixed_r_asa != None:
            for bond in self.isopeptide_bonds:
                bond.r_asa = self.fixed_r_asa
        else:
            self._calc_rasa()
        # Make prediction with linear regression
        self._infer()
        # Sort based on probability
        self.isopeptide_bonds.sort(key=lambda x: (x.probability, x.protein_name), reverse=True)
        # Infer type of bond
        self._infer_type()

    def print_tabular(self):
        """
        
            Print isopeptide bonds in a tabular format

            Example:
                >>> from isopeptor.isopeptide import Isopeptide
                >>> i = Isopeptide("tests/data/test_structures", distance=1.5, fixed_r_asa=0.1)
                >>> i.predict()
                >>> i.print_tabular()
                protein_name        probability     chain   r1_bond r_cat   r2_bond r1_bond_name    r_cat_name      r2_bond_name    bond_type       rmsd    r_asa   template             
                8beg                0.991           A       590     636     729     LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_590_636_729   
                8beg                0.991           A       756     806     894     LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_756_806_894   
                8beg                0.991           A       922     973     1049    LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_922_973_1049  
                8beg                0.991           A       1076    1123    1211    LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_1076_1123_1211
                5dz9                0.991           A       556     606     703     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_3_53_150      
                5dz9                0.991           A       730     776     861     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_177_223_308   
                4z1p                0.991           A       3       53      150     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_3_53_150      
                4z1p                0.991           A       177     223     308     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_177_223_308   
                7woi                0.909           B       57      158     195     LYS             GLU             ASN             CnaB-like       0.314   0.1     5j4m_A_47_139_172    
                1amx                0.882           A       176     209     293     LYS             ASP             ASN             CnaA-like       0.353   0.1     2f68_X_176_209_293   
                7woi                0.875           A       203     246     318     LYS             ASP             ASN             CnaA-like       0.363   0.1     4hss_A_187_224_299   
                7woi                0.838           B       355     435     466     LYS             GLU             ASN             CnaB-like       0.403   0.1     8f70_A_299_386_437   
                6to1_af             0.607           A       13      334     420     LYS             ASP             ASN             CnaA-like       0.565   0.1     5mkc_A_191_600_695   

        """
        if not self.geometry_calculated:
            headers = [
            "protein_name", "probability", "chain", "r1_bond", "r_cat", "r2_bond",
            "r1_bond_name", "r_cat_name", "r2_bond_name", "bond_type",
            "rmsd", "r_asa", "template"
            ]
            # Print the header row using formatted widths
            column_widths = []
            for header in headers:
                # Find the maximum length among the bond attributes for the current header
                max_bond_length = 0
                for bond in self.isopeptide_bonds:
                    bond_value_length = len(str(getattr(bond, header)))
                    if bond_value_length > max_bond_length:
                        max_bond_length = bond_value_length
                # Compare it with the length of the header and append the maximum to column_widths
                column_widths.append(max(len(header), max_bond_length))
            formatted_header = "\t".join(f"{header:<{column_widths[i]}}" for i, header in enumerate(headers))
            print(formatted_header)
            
            if len(self.isopeptide_bonds) > 0:
                column_widths = [max(len(header), max(len(str(getattr(bond, header))) for bond in self.isopeptide_bonds)) for header in headers]
                for bond in self.isopeptide_bonds:
                    row = [
                        bond.protein_name, bond.probability, bond.chain, bond.r1_bond, bond.r_cat, bond.r2_bond, 
                        bond.r1_bond_name, bond.r_cat_name, bond.r2_bond_name, bond.bond_type,
                        bond.rmsd, bond.r_asa, bond.template
                    ]
                    formatted_row = "\t".join(f"{str(item):<{column_widths[i]}}" for i, item in enumerate(row))
                    print(formatted_row)
        else:
            headers = [
                "protein_name", "probability", "chain", "r1_bond", "r_cat", "r2_bond",
                "r1_bond_name", "r_cat_name", "r2_bond_name", "bond_type",
                "rmsd", "r_asa", "template", "bond_length", "bond_length_zscore", "bond_length_allowed",
                "pseudo_phi", "pseudo_psi", "pseudo_omega", "phi_psi_likelihood", "phi_psi_allowed", "omega_psi_likelihood",
                "omega_psi_allowed",  "omega_phi_likelihood", "omega_phi_allowed"
            ]
            # Print the header row using formatted widths
            column_widths = []
            for header in headers:
                # Find the maximum length among the bond attributes for the current header
                max_bond_length = 0
                for bond in self.isopeptide_bonds:
                    bond_value_length = len(str(getattr(bond, header)))
                    if bond_value_length > max_bond_length:
                        max_bond_length = bond_value_length
                # Compare it with the length of the header and append the maximum to column_widths
                column_widths.append(max(len(header), max_bond_length))
                
            formatted_header = "\t".join(f"{header:<{column_widths[i]}}" for i, header in enumerate(headers))
            print(formatted_header)
            if len(self.isopeptide_bonds) > 0:                
                # Print each data row using the same formatting
                for bond in self.isopeptide_bonds:
                    row = [
                        bond.protein_name, bond.probability, bond.chain, bond.r1_bond, bond.r_cat, bond.r2_bond, 
                        bond.r1_bond_name, bond.r_cat_name, bond.r2_bond_name, bond.bond_type,
                        bond.rmsd, bond.r_asa, bond.template, bond.bond_length, bond.bond_length_zscore, bond.bond_length_allowed,
                        bond.pseudo_phi, bond.pseudo_psi, bond.pseudo_omega, bond.phi_psi_likelihood, bond.phi_psi_allowed, bond.omega_psi_likelihood,
                        bond.omega_psi_allowed, bond.omega_phi_likelihood, bond.omega_phi_allowed
                    ]
                    formatted_row = "\t".join(f"{str(item):<{column_widths[i]}}" for i, item in enumerate(row))
                    print(formatted_row)
    
    def save_csv(self, output_table: str = "results.csv") -> None:
        """
        
            Save isopeptide bond results in .csv format.
            
            Args:
                output_table (str): path to output table
            
            Example:
                >>> from isopeptor.isopeptide import Isopeptide
                >>> i = Isopeptide("tests/data/test_structures", distance=1.5, fixed_r_asa=0.1)
                >>> i.predict()
                >>> i.save_csv()
                >>> # Or with geometry evaluation
                >>> i.get_geometry()
                >>> i.save_csv()
        
        """
        with open(output_table, "wt") as fh:
            if not self.geometry_calculated:
                headers = [
                "protein_name", "probability", "chain", "r1_bond", "r_cat", "r2_bond",
                "r1_bond_name", "r_cat_name", "r2_bond_name", "bond_type",
                "rmsd", "r_asa", "template"
                ]
                fh.write(",".join(headers)+"\n")
                if len(self.isopeptide_bonds) > 0:
                    for bond in self.isopeptide_bonds:
                        row = [
                            bond.protein_name, bond.probability, bond.chain, bond.r1_bond, 
                            bond.r_cat, bond.r2_bond, 
                            bond.r1_bond_name, bond.r_cat_name, bond.r2_bond_name, bond.bond_type,
                            bond.rmsd, bond.r_asa, bond.template
                        ]
                        row = [str(i) for i in row]
                        formatted_row = ",".join(row)
                        fh.write(formatted_row+"\n")
            else:
                headers = [
                    "protein_name", "probability", "chain", "r1_bond", "r_cat", "r2_bond",
                    "r1_bond_name", "r_cat_name", "r2_bond_name", "bond_type",
                    "rmsd", "r_asa", "template", "bond_length", "bond_length_zscore", "bond_length_allowed",
                    "pseudo_phi", "pseudo_psi", "pseudo_omega", "phi_psi_likelihood", "phi_psi_allowed", "omega_psi_likelihood",
                    "omega_psi_allowed",  "omega_phi_likelihood", "omega_phi_allowed"
                ]

                if len(self.isopeptide_bonds) > 0:
                    
                    formatted_header = ",".join(headers)
                    fh.write(formatted_header+"\n")
                    for bond in self.isopeptide_bonds:
                        row = [
                            bond.protein_name, bond.probability, bond.chain, 
                            bond.r1_bond, bond.r_cat, bond.r2_bond, 
                            bond.r1_bond_name, bond.r_cat_name, bond.r2_bond_name, bond.bond_type,
                            bond.rmsd, bond.r_asa, bond.template, bond.bond_length, 
                            bond.bond_length_zscore, bond.bond_length_allowed,
                            bond.pseudo_phi, bond.pseudo_psi, bond.pseudo_omega, 
                            bond.phi_psi_likelihood, bond.phi_psi_allowed, bond.omega_psi_likelihood,
                            bond.omega_psi_allowed, bond.omega_phi_likelihood, bond.omega_phi_allowed
                        ]
                        row = [str(i) for i in row]
                        formatted_row = ",".join(row)
                        fh.write(formatted_row+"\n")

    def get_geometry(self):
        """
        
            Get geometry measures (bond length and dihedral angles and map to existing distribution of measures from 
            the database of PDB derived structures).

            Example:
                >>> from isopeptor.isopeptide import Isopeptide
                >>> i = Isopeptide("tests/data/test_structures", distance=1.5, fixed_r_asa=0.1)
                >>> i.predict()
                >>> i.get_geometry()
                >>> i.print_tabular()
                protein_name        probability     chain   r1_bond r_cat   r2_bond r1_bond_name    r_cat_name      r2_bond_name    bond_type       rmsd    r_asa   template                bond_length    bond_length_zscore       bond_length_allowed     pseudo_phi      pseudo_psi      pseudo_omega    phi_psi_likelihood      phi_psi_allowed omega_psi_likelihood    omega_psi_allowed       omega_phi_likelihood    omega_phi_allowed
                8beg                0.991           A       590     636     729     LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_590_636_729      1.33           0.024                    True                    92.305          -123.286        -0.032          -8.706                  True            -7.801                  True                    -8.797                  True             
                8beg                0.991           A       756     806     894     LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_756_806_894      1.328          0.0                      True                    -114.423        -130.136        44.142          -10.503                 False           -10.472                 True                    -10.503                 False            
                8beg                0.991           A       922     973     1049    LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_922_973_1049     1.333          0.061                    True                    70.944          -131.258        17.005          -9.737                  True            -8.997                  True                    -9.882                  True             
                8beg                0.991           A       1076    1123    1211    LYS             ASP             ASN             CnaA-like       0.0     0.1     8beg_A_1076_1123_1211   1.341          0.159                    True                    102.317         -124.836        2.258           -8.286                  True            -7.882                  True                    -8.233                  True             
                5dz9                0.991           A       556     606     703     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_3_53_150         1.291          -0.451                   True                    109.407         -112.845        -8.356          -7.72                   True            -7.722                  True                    -7.884                  True             
                5dz9                0.991           A       730     776     861     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_177_223_308      1.332          0.049                    True                    97.334          -122.501        6.356           -8.391                  True            -7.892                  True                    -8.633                  True             
                4z1p                0.991           A       3       53      150     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_3_53_150         1.291          -0.451                   True                    109.407         -112.845        -8.356          -7.72                   True            -7.722                  True                    -7.884                  True             
                4z1p                0.991           A       177     223     308     LYS             ASP             ASN             CnaA-like       0.0     0.1     4z1p_A_177_223_308      1.332          0.049                    True                    97.334          -122.501        6.356           -8.391                  True            -7.892                  True                    -8.633                  True             
                7woi                0.909           B       57      158     195     LYS             GLU             ASN             CnaB-like       0.314   0.1     5j4m_A_47_139_172       1.312          -0.195                   True                    -148.85         71.47           179.899         -9.376                  True            -9.126                  True                    -9.675                  True             
                1amx                0.882           A       176     209     293     LYS             ASP             ASN             CnaA-like       0.353   0.1     2f68_X_176_209_293      3.345          24.598                   False                   96.887          -147.455        66.383          -10.918                 False           -14.477                 False                   -25.835                 False            
                7woi                0.875           A       203     246     318     LYS             ASP             ASN             CnaA-like       0.363   0.1     4hss_A_187_224_299      1.336          0.098                    True                    136.992         -158.356        4.495           -14.872                 False           -12.529                 False                   -9.337                  True             
                7woi                0.838           B       355     435     466     LYS             GLU             ASN             CnaB-like       0.403   0.1     8f70_A_299_386_437      1.36           0.39                     True                    71.106          149.083         171.017         -11.508                 False           -9.75                   True                    -11.537                 False            
                6to1_af             0.607           A       13      334     420     LYS             ASP             ASN             CnaA-like       0.565   0.1     5mkc_A_191_600_695      2.928          19.512                   False                   83.132          -155.408        131.186         -13.496                 False           -51.581                 False                   -15.775                 False
        
        """
        self.geometry_calculated = True
        bonds = self.isopeptide_bonds
        for struct_file in set([b.struct_file for b in bonds]):
            structure = get_structure(struct_file)
            for bond in [b for b in bonds if b.struct_file == struct_file]:
                # Get bond lenght and stats
                bond_length, bond_length_zscore, bond_length_allowed = get_bond_stats(structure, bond.chain, bond.r1_bond, 
                                bond.r2_bond, bond.r2_bond_name)
                
                # Get dihedrals and stats
                angle_stats = get_dihedral_angles_stats(structure, bond.chain, bond.r1_bond, 
                                bond.r2_bond, bond.r2_bond_name, bond.bond_type)
                pseudo_omega, pseudo_phi, pseudo_psi, phi_psi_likelihood, omega_psi_likelihood, omega_phi_likelihood, phi_psi_allowed, omega_psi_allowed, omega_phi_allowed = angle_stats

                # Load values into the bond instance
                bond.bond_length = bond_length
                bond.bond_length_zscore = bond_length_zscore
                bond.bond_length_allowed = bond_length_allowed
                bond.pseudo_omega = pseudo_omega
                bond.pseudo_phi = pseudo_phi
                bond.pseudo_psi = pseudo_psi
                bond.phi_psi_likelihood = phi_psi_likelihood
                bond.omega_psi_likelihood = omega_psi_likelihood
                bond.omega_phi_likelihood = omega_phi_likelihood
                bond.phi_psi_allowed = phi_psi_allowed
                bond.omega_psi_allowed = omega_psi_allowed
                bond.omega_phi_allowed = omega_phi_allowed

    def _load_hits(self):
        """

            Load pyjess._jess.Hit as list of bondElement in self.isopeptide_bonds.

            Raises:
                ValueError if number of residues found is not expected.

        """
        for hit in self.jess_hits:
            template, rmsd, struct_file, atoms = hit.template.id, hit.rmsd, hit.molecule.id, hit.atoms()
            protein_name = os.path.basename(struct_file).replace(".pdb", "").replace(".cif", "")
            residues, residue_names = [], []
            for atom in atoms:
                if atom.residue_number not in residues:
                    residues.append(atom.residue_number)
                    residue_names.append(atom.residue_name)
            chain = atom.chain_id

            if len(residues) == 3:
                self.isopeptide_bonds.append(
                    BondElement(
                                struct_file, protein_name, round(rmsd, 3), template, chain, 
                                residues[0], residues[1], residues[2],
                                residue_names[0], residue_names[1], residue_names[2]
                    )
                )
            else:
                raise ValueError(f"Found {len(residues)} residues.")

    def _reduce_redundant(self):
        """

            Reduces redundant isopeptide bond predictions by keeping prediction with lower RMSD.

        """
        bonds = self.isopeptide_bonds
        grouped_bonds = {}
        
        for bond in bonds:
            key = (bond.protein_name, bond.r1_bond, bond.r_cat, bond.r2_bond)
            if key not in grouped_bonds or bond.rmsd < grouped_bonds[key].rmsd:
                grouped_bonds[key] = bond

        # Keep only the best bond for each group
        self.isopeptide_bonds = list(grouped_bonds.values())

    def _calc_rasa(self):
        """

            Calculate r_asa for every isopeptide bond in every structure.

        """
        bonds = self.isopeptide_bonds
        for struct_file in set([b.struct_file for b in bonds]):
            # Get asa and structure atom array pdb file
            structure_sasa, structure = get_structure_asa(struct_file)
            # Get asa of isopeptide residues
            for bond in [b for b in bonds if b.struct_file == struct_file]:
                isopep_residues = [bond.r1_bond, bond.r_cat, bond.r2_bond]
                isopep_residue_names = [bond.r1_bond_name, bond.r_cat_name, bond.r2_bond_name]
                tmp_r_asa = 0
                for res_id, res_name in zip(isopep_residues, isopep_residue_names):
                    # Handle the case of two res bb
                    if res_name == None:
                        continue
                    res_indeces = [i for i, atom in enumerate(structure) if atom.res_id == res_id and atom.chain_id == bond.chain]
                    #res_name = structure[res_indeces[0]].res_name
                    if res_name not in MAX_ASA["rost_sander"].keys():
                        raise ValueError(f"{res_name} not in {MAX_ASA['rost_sander'].keys()}")
                    # Normalise ASA by residue surface area
                    r_asa = sum([structure_sasa[i] for i in res_indeces]) / MAX_ASA["rost_sander"][res_name]
                    tmp_r_asa += r_asa
                bond.r_asa = round(tmp_r_asa / 3, 3)
    
    def _infer(self):
        """

            Infer presence of isoeptide bond using logistic regression model.

        """
        for bond in self.isopeptide_bonds:
            bond.probability = predict(bond.rmsd, bond.r_asa)

    def _infer_type(self):
        """

            Infer isopeptide bond type (CnaA/B-like).

        """
        for bond in self.isopeptide_bonds:
            bond.bond_type = BOND_TYPE.get(bond.template, None)

if __name__ == "__main__":
    import doctest
    doctest.testmod(report=True, optionflags=doctest.NORMALIZE_WHITESPACE)